/***************************************************************************/
/* The Minstrel's Song is (c) 1998-2003 by Jonathan Hayward.  See the file */
/* lib/copyrigh.txt for a copyright and license statement.                 */
/***************************************************************************/

#include <math.h>
#include <string.h>
#include "tms.h"

long def_sec(struct small_creature *c, long seconds)
    {
    return 1+ (long) (((double) seconds*UNITS_PER_SECOND)/(pow(2.0,
    ((double) (c->sp)/10.0))));
    }

struct small_creature *loc_creature(struct small_creature *c)
    {
    long x, y;
    loc_xy(c, &x, &y);
    if (current_level->space[x][y].creatures != NULL)
        return current_level->space[x][y].creatures->current;
    else
        return NULL;
    }

void flag_set(long *target, long flags)
    {
    *target=(*target | flags); /* check to make sure bitwise */
    }

void flag_reset(long *target, long flags)
    {
    *target=(*target & (~flags)); /* check to make sure bitwise */
    }

long flag_check(long target, long my_flag)
    {
    return (!(0 == (target & my_flag)));
    }

long skill_check(struct small_creature *c, long skill, long difficulty, long
  flags)

/* contains critical success/failure */

    {
    /* -n nth degree critical failure, 0 failure, 1 success, n+1
    nth degree critical success */ /* check this */
    long index, min, result;
    double ratio,temp;
    index=skill-difficulty;
    if (!(flag_check(flags, SKILL_CHECK_NO_SIGHT_ADJUST)))
        if (!(can_see(c, c->x, c->y)))
            {
            if (c->extended != NULL)
                {
                index -= ONE_NOTCH * max(0, (2 - skill_check(c,
                  c->extended->skills->av[BLIND_ACTION], MODERATE,
                  (SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE * (flag_check(
                  flags, SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE) != 0)) |
                  SKILL_CHECK_NO_SIGHT_ADJUST)));
                }
            else
                {
                index -= ONE_NOTCH * max(0, (2 - skill_check(c,
                  VERY_DIFFICULT, MODERATE,
                  (SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE * (flag_check(
                  flags, SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE) != 0)) |
                  SKILL_CHECK_NO_SIGHT_ADJUST)));
                }
            }
    temp=((double) trandom() + 1.0)/32769.0; /* avoiding division by zero... */
    ratio=-temp/(temp-1.0);
    min=(long) (0.5+(log(ratio)/(log(2.0)/10.0)));
    result = (index-min)/40 + 1;
    if ((result > 1) && (flag_check(flags,
      SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE)))
        gain_experience(c, (result - 1) * (result - 1));
    return result;
    }

long can_see(struct small_creature *c, long x, long y)
    {
    long i, line_of_sight, total, visible, c_x, c_y;
    if (c->extended != NULL)
        if (skill_check(c, c->extended->skills->av[KEEN_EYESIGHT], DIFFICULT,
          SKILL_CHECK_NO_SIGHT_ADJUST) > 0)
            if (can_see_badly(c, x, y))
                return TRUE;
    total=abs(c->x-x)+abs(c->y-y);
    line_of_sight=1;
    c_x = c->x;
    c_y = c->y;
    visible=((flag_check(current_level->space[x][y].flags, CURR_ILLUMINATED))
      || ((flag_check(current_level->space[x][y].flags, DIMLY_LIT)) &&
      (abs(c->x - x) <= 1) && (abs(c->y - y) <= 1)));
    if ((!(visible)) || (flag_check(c->flags, CANT_SEE)))
        return 0;
    if (!((abs(c->x-x) <= 1) && (abs(c->y-y) <= 1)))
        for(i=0; ((i < total) && line_of_sight); ++i)
            line_of_sight&=((
            current_level->space[c_x+i*(x-c_x)/total][c_y+i*(y-c_y)
            /total].flags&TRANSPARENT) && (!(
            current_level->space[c_x+i*(x-c_x)/total][c_y+i*(y-c_y)
            /total].flags&S_CLOAKED_BY_MIST)));
    return line_of_sight;
    }

long can_see_badly(struct small_creature *c, long x, long y)
    {
    long i, line_of_sight, total, visible, c_x, c_y;
    total=abs(c->x-x)+abs(c->y-y);
    line_of_sight=1;
    c_x = c->x;
    c_y = c->y;
    visible=(flag_check(current_level->space[x][y].flags, DIMLY_LIT) ||
      flag_check(current_level->space[x][y].flags, CURR_ILLUMINATED));
    if ((!(visible)) || (flag_check(c->flags, CANT_SEE)))
        return 0;
    /* We can see badly through one space of fog.  */
    if (!((abs(c->x-x) <= 2) && (abs(c->y-y) <= 2)))
        for(i=0; ((i < total) && line_of_sight); ++i)
            line_of_sight&=((
            current_level->space[c_x+i*(x-c_x)/total][c_y+i*(y-c_y)
            /total].flags&TRANSPARENT) && (!(
            current_level->space[c_x+i*(x-c_x)/total][c_y+i*(y-c_y)
            /total].flags&S_CLOAKED_BY_MIST)));
    else
        for(i=0; ((i < total) && line_of_sight); ++i)
            if (!((abs(c_x-x) <= 1) && (abs(c_y-y) <= 1)))
                line_of_sight=(line_of_sight && (flag_check(
                current_level->space[c_x+i*(x-c_x)/total][c_y+i*(y-c_y)
                /total].flags,TRANSPARENT)));
    return line_of_sight;
    }

long clear_path(long x1, long y1, long x2, long y2)
    {
    long i, is_clear_path, total;
    total=abs(x1 - x2)+abs(y1 - y2);
    is_clear_path=1;
    for(i=0; i < total; ++i)
        if (!((abs(x1 - x2) < 1) && (abs(x1 - x2) < 1)))
            is_clear_path=(is_clear_path && (!(flag_check(
            current_level->space[x1+i*(x2 - x1)/total][y1+i*(y2 - y1)
            /total].flags,OCCUPIED))));
    return is_clear_path;
    }

long can_see_water(struct small_creature *c)
    {
    long i, x, y, line_of_sight, total, water_seen;
    water_seen = 0;
    for(x=0;x<MAX_X;++x)
        for(y=0;y<MAX_Y;++y)
            {
            total=abs(c->x-x)+abs(c->y-y);
            line_of_sight=1;
            for(i=0; i < total; ++i)
                /* We can usually see what's next to us */
                if (!((abs(c->x-x) <= 1) && (abs(c->y-y) <= 1)))
                    line_of_sight=(line_of_sight && (flag_check(
                      current_level->space[c->x+i*(x-c->x)/total]
                      [c->y+i*(y-c->y) /total].flags,TRANSPARENT) &&
                      (((abs(c->x-x) <= 2) && (abs(c->y-y) <= 2)) ||
                      (!(flag_check( current_level->space
                      [c->x+i*(x-c->x)/total][c->y+i*(y-c->y)
                      /total].flags,S_CLOAKED_BY_MIST))))));
            water_seen=(water_seen || (line_of_sight &&
              (flag_check(current_level->space[x][y].flags,
              CURR_ILLUMINATED|IS_WATER)) &&
              (!(flag_check(current_level->space[x][y].flags, SALT_WATER)))));
            }
    return (water_seen && (!flag_check(c->flags, CANT_SEE)));
    }

long my_exp(long base_points)
    {
    return (long) (pow(exp(log(2.0)/10.0),(double) base_points));
    }

long my_log(long base_points)
    {
    return (long) (log((double) base_points) / (log(2.0)/10.0));
    }

struct small_item *get_item(struct small_creature *c, long flags)
    {
    char player_choice;
    long temp;
    struct small_item *result;
    struct ilist **target_ilist;
    temp = 'a';
    if (!(c->extended == NULL))
        {
        if (flag_check(flags, EXTRACT_PROMPT_BOTH_PACK_AND_EQUIPMENT))
            {
            pmsg("Take item from (p)ack, or (e)quipment currently in use?");
            pmsg("('?' to list all inventory in pack and equipment.)");
            switch(tgetc())
                {
                case 'p':
                    target_ilist = &c->extended->pack;
                break;
                case 'e':
                    target_ilist = &c->extended->wear_wield;
                break;
                case '?':
                    print_inventory(c);
                    return get_item(c, flags);
                break;
                default:
                    pmsg("Aborting...");
                    return NULL;
                break;
                }
            }
        else
            {
            target_ilist = &c->extended->pack;
            }
        pmsg("Please enter the item's slot letter ('?' to list).");
        for(player_choice = tgetc(); player_choice == '?'; player_choice =
          tgetc())
            {
            if (player_choice == '?')
                {
                print_inventory(c);
                pmsg("Please enter the item's slot letter ('?' to list).");
                }
            }
        result = extract_item(target_ilist, 'a', &temp,
          player_choice, EXTRACT_RECURSIVE);
        if (flag_check(flags, EXTRACT_REMOVE_ITEM))
            find_and_remove_item(result, target_ilist);
        return result;
        }
    else
        return NULL;
    }

long confirm_pos(void)
    {
    switch(trandom()%2)
        {
        case 0: pmsg("Correct (y/n)?"); break;
        case 1: pmsg("Right (y/n)?"); break;
        }
    return confirm_silent();
    }

long confirm_neg(void)
    {
    switch(trandom()%3)
        {
        case 0: pmsg("Are you sure (y/n)?"); break;
        case 1: pmsg("Do you really want to do that (y/n)?"); break;
        case 2: pmsg("Do you really mean it (y/n)?"); break;
        }
    return confirm_silent();
    }

long confirm_silent(void)
    {
    long result;
    result = confirm_silent_internal();
    if (result)
        pmsg("Yes.");
    else
        pmsg("No.");
    return result;
    }

long confirm_silent_internal(void)
    {
    char resp;
    long result;
    for(resp=tgetc(); (!(resp=='y') && !(resp=='n') && (!(resp == ESCAPE)));
      resp=tgetc());
    result = (resp=='y');
    return result;
    }

long confirm(void)
    {
    long result;
    if (flag_check(game_variables->flags, NEGATIVE_CONFIRM))
        result = confirm_neg();
    else
        result = confirm_pos();
    return result;
    }

long confirm_maybe(void) /* If the appropriate game variable is set.  */
    {
    if (flag_check(game_variables->flags, CONFIRM_WHEN_IN_DOUBT))
        return confirm_pos();
    else
        return 1;
    }

long max(long a, long b)
    {
    if (a>b) return a; else return b;
    }

long min(long a, long b)
    {
    if (a<b) return a; else return b;
    }

void marker(char *message)
    {
#ifdef DEBUG
    /*
    printf("Reached here!%c\n%c\nNote: %s%c\n%c\n", 13, 13, message, 13, 13);
    */
    /* putchar(7); */
    /*
    fflush(stdout);
    */
    if (flag_check(game_variables->flags, DEBUG_MODE))
        pmsg(message);
#endif
    }

long chances(long a, long b)
    {
    return (trandom()%b < a);
    }

void description(struct small_item *i, char result[])
    {
    if (flag_check(i->flags, IDENTIFIED))
        strncpy(result, orig_items[i->identity].identified, 255);
    else
        strncpy(result, orig_items[i->identity].nonidentified, 255);
    result[255] = '\0';
    }

void select_random_result(long table[NUM_DIST_SLOTS][3], long *selected_type,
  long *selected_other)
    {
    long i, total, desired, so_far, result[2];
    total = so_far = result[0] = result[1] = 0;
    for(i=0; i < NUM_DIST_SLOTS; ++i)
        total += table[i][2];
    desired = trandom() % total + 1;
    for(i=0; i < NUM_DIST_SLOTS; ++i)
        {
        so_far += table[i][2];
        if ((so_far - table[i][2] < desired) && (desired <= so_far))
            {
            *selected_type=table[i][0];
            *selected_other=table[i][2];
            }
        }
    }

/* Find the next character in an order that goes a-z, A-Z, 1-9, and then ?.*/

long successor(long letter)
    {
    long result;
    if (letter == 'z')
        result='A';
    else if (letter == 'Z')
        result='1';
    else if ((letter == '9') || (letter == '?'))
        result='?';
    else
        result=letter+1;
    return result;
    }

long predecessor(long letter)
    {
    long result;
    if (letter == 'A')
        result='z';
    else if (letter == '1')
        result='Z';
    else if (letter == '?')
        result=-1; /* Special case; '?' denotes left over slots. */
    else
        result=letter-1;
    return result;
    }

void random_select(long distribution[NUM_DIST_SLOTS][3], long *result, long
  *other)
    {
    long i, my_total, chosen, so_far;
    my_total = 0;
    so_far = 0;
    for(i=0;i<NUM_DIST_SLOTS;++i)
        my_total += distribution[i][1];
    chosen = trandom() % my_total + 1;
    for(i=0;i<NUM_DIST_SLOTS;++i)
        {
        so_far += distribution[i][1];
        if ((so_far - distribution[i][1] < chosen) && (chosen <= so_far))
            {
            *result = distribution[i][0];
            *other = distribution[i][2];
            }
        }
    }

void two_random_select(long distribution[NUM_DIST_SLOTS][3], long *result1,
  long *result2)
    {
    long i, my_total, chosen, so_far;
  my_total = 0;
  so_far = 0;
  for(i=0;i<NUM_DIST_SLOTS;++i)
    my_total += distribution[i][1];
  chosen = trandom() % my_total + 1;
  for(i=0;i<NUM_DIST_SLOTS;++i)
    {
    so_far += distribution[i][1];
    if ((so_far - distribution[i][1] < chosen) && (chosen <= so_far))
            *result1 = distribution[i][0];
    }
  my_total = 0;
  so_far = 0;
  for(i=0;i<NUM_DIST_SLOTS;++i)
    my_total += distribution[i][2];
  chosen = trandom() % my_total + 1;
  for(i=0;i<NUM_DIST_SLOTS;++i)
    {
    so_far += distribution[i][2];
    if ((so_far - distribution[i][2] < chosen) && (chosen <= so_far))
            *result2 = distribution[i][0];
    }
    }

void remove_creature(struct level *l, struct small_creature *c)
    {
    flag_reset(&l->space[c->x][c->y].flags, OCCUPIED);
    remove_creature_from_clist(c, &l->space[c->x][c->y].creatures);
    remove_creature_from_clist(c, &l->creatures);
    }

void add_creature(struct level *l, struct small_creature *c, long x, long y)
    {
    /* hostile monsters attack anybody foreign that enters the level.  */
    hostility_check(l, c);
    no_hostile_add_creature(l, c, x, y);
    }

void no_hostile_add_creature(struct level *l, struct small_creature *c, long x,
  long y)
    {
    long i, j, found_x, found_y;
    found_x = x;
    found_y = y;
    if (!(flag_check(l->flags, OUTDOORS)))
        land_from_flying(c);
    /* If the given space is occupied, look for one next to it.  */
    for (i=max(0, x - 1); i < min(x + 2, MAX_X); ++i)
        for (j = max(0, y - 1); j < min(y + 2, MAX_Y); ++j)
            if ((flag_check(l->space[i][j].flags, CAN_OCCUPY)) &&
                (!(flag_check(l->space[i][j].flags, OCCUPIED))))
                {
                found_x = i;
                found_y = j;
                }
    if ((flag_check(l->space[x][y].flags, CAN_OCCUPY)) &&
        (!(flag_check(l->space[x][y].flags, OCCUPIED))))
        {
        found_x = x;
        found_y = y;
        }
    c->x = found_x;
    c->y = found_y;
    flag_set(&l->space[found_x][found_y].flags, OCCUPIED);
    append_creature_to_clist(c, &l->creatures);
    append_creature_to_clist(c, &l->space[found_x][found_y].creatures);
    if (c == player)
        {
        game_variables->display_center_x=c->x;
        game_variables->display_center_y=c->y;
        }
    if (flag_check(c->flags, OBSCURE_VISION))
        {
        mist_circle(l, c->x, c->y, 2);
        }
    flag_reset(&c->flags, SHIFT_MOVE);
    flag_reset(&c->flags, CONTROL_MOVE);
    }

/* Now a box metric; could change to taxicab metric or something else */
/* similar.  */

long distance(long x1, long y1, long x2, long y2)
    {
    return max(abs(x1 - x2), abs(y1 - y2));
    }

struct small_creature *nearest_hostile(struct small_creature *c)

/* Not just the nearest creature to which it is hostile.  It will first go to */
/* the nearest creature to which it is hostile and to which there is a clear */
/* path, then to the nearest creature to which it is hostile and which it can */
/* see, and then to the nearest creature to which it is hostile.  */

    {
    long creature_clear_path, creature_can_see, best_x, best_y, is_candidate;
    struct small_creature *current_best;
    struct clist *candidate;
    best_x = best_y = 32767;
    creature_clear_path = creature_can_see = is_candidate = 0;
    current_best = NULL;
    for(candidate = c->hostile; (candidate!=NULL); candidate = candidate->next)
        {
        is_candidate = FALSE;
        if (!(current_best==NULL))
            {
            is_candidate = TRUE;
            is_candidate &= (is_on_level(candidate->current, current_level));
            if (creature_clear_path)
                is_candidate = (is_candidate && (clear_path(c->x, c->y,
                  candidate->current->x, candidate->current->y)));
            else if (creature_can_see)
                is_candidate = (is_candidate && (can_see(c,
                  candidate->current->x, candidate->current->y)));
            is_candidate = (is_candidate && (distance(c->x, c->y,
              candidate->current->x, candidate->current->y) <= distance(c->x,
              c->y, best_x, best_y)));
            if (is_candidate)
                {
                best_x = candidate->current->x;
                best_y = candidate->current->y;
                current_best = candidate->current;
                if (can_see(c, candidate->current->x, candidate->current->y))
                    creature_can_see = 1;
                if (clear_path(c->x, c->y, candidate->current->x,
                  candidate->current->y))
                    creature_clear_path = 1;
                }
            }
        else
            {
            current_best = candidate->current;
            if (can_see(c, candidate->current->x, candidate->current->y))
                creature_can_see = 1;
            if (clear_path(c->x, c->y, candidate->current->x,
              candidate->current->y))
                creature_clear_path = 1;
            }
        }
    return current_best;
    }

void find_nearest_water(struct small_creature *c, long *found_x, long *found_y)
    {
    long i, j, best_x, best_y;
    best_x = best_y = 0;
    for(i=0;i<MAX_X;++i)
        for(j=0;j<MAX_Y;++j)
            if ((distance(c->x, c->y, i, j) <= distance(c->x, c->y, best_x,
              best_y))
              && (can_see(c, i, j)) && (flag_check(current_level->space[i][j].
              flags, IS_WATER)) && (!(flag_check(current_level->space[i][j].
              flags, SALT_WATER))))
                {
                best_x = i;
                best_y = j;
                }
    *found_x = best_x;
    *found_y = best_y;
    }

/* Is c1 hostile to c2? */

long is_hostile(struct small_creature *c1, struct small_creature *c2)
    {
    long result;
    struct clist *clcurr;
    result = FALSE;
    if (!(c1->hostile == NULL))
        for(clcurr=c1->hostile;(!(clcurr==NULL));clcurr=clcurr->next)
            if (clcurr->current == c2)
                result = TRUE;
    return result;
    }

long is_following(struct small_creature *c1, struct small_creature *c2)
    {
    long result1, result2;
    struct clist *clcurr;
    result1 = result2 = FALSE;
    if (!(c1->follow == NULL))
        for(clcurr=c1->follow;(!(clcurr==NULL));clcurr=clcurr->next)
            if (clcurr->current == c2)
                result1 = TRUE;
    if (!(c2->followed_by == NULL))
        for(clcurr=c2->followed_by;(!(clcurr==NULL));clcurr=clcurr->next)
            if (clcurr->current == c1)
                result2 = TRUE;
    return (result1 && result2);
    }

void reset_hostility(struct small_creature *c1, struct small_creature *c2)
    {
    remove_creature_from_clist(c2, &c1->hostile);
    }

void reset_follow(struct small_creature *c1, struct small_creature *c2)
    {
    remove_creature_from_clist(c2, &c1->follow);
    remove_creature_from_clist(c1, &c2->followed_by);
    }

/* We return success or failure; failure is in case the player hits escape */
long loc_xy(struct small_creature *c, long *result_x, long *result_y)
    {
    long curr_x, curr_y;
    char command;
    pmsg("Targeting...  hit \'?\' for help.");
    curr_x = c->x;
    curr_y = c->y;
    move(((curr_y - game_variables->display_center_y) + 6 + (LINES-10)/2),
    ((curr_x - game_variables->display_center_x) + 13 + (COLS-16)/2));
    refresh();
    for(command=tgetc();(!((command=='.') || (command == ESCAPE)));
      command=tgetc())
        {
        switch(command)
            {
            case LEFT: case SECOND_LEFT: --curr_x; break;
            case DOWN: case SECOND_DOWN: ++curr_y; break;
            case UP: case SECOND_UP: --curr_y; break;
            case RIGHT: case SECOND_RIGHT: ++curr_x; break;
            case UP_LEFT: case SECOND_UP_LEFT: --curr_x; --curr_y; break;
            case UP_RIGHT: case SECOND_UP_RIGHT: ++curr_x; --curr_y; break;
            case DOWN_LEFT: case SECOND_DOWN_LEFT: --curr_x; ++curr_y; break;
            case DOWN_RIGHT: case SECOND_DOWN_RIGHT: ++curr_x; ++curr_y; break;
            case '?':
                pmsg("Use the motion commands to move the cursor about, until it has reached the");
                pmsg("location you wish to select.  Then hit \'.\'.  If you wish to abort, hit");
                pmsg("Escape.");
            break;
            case ESCAPE: pmsg("Aborted."); break;
            }
        /* If the player has moved out of window, return to the window.  I am */
        /* for now choosing not to do window scrolling, as I found that to be */
        /* an annoying attribute of omega.  */
        curr_x = max(curr_x, game_variables->display_center_x - (COLS - 16)/2);
        curr_x = min(curr_x, game_variables->display_center_x - 1 + (COLS - 16)
          /2);
        curr_y = max(curr_y, game_variables->display_center_y - (LINES - 10)/2);
        curr_y = min(curr_y, game_variables->display_center_y - 1 + (LINES - 10)
          /2);
        move(((curr_y - game_variables->display_center_y) + 6 + (LINES-10)/2),
        ((curr_x - game_variables->display_center_x) + 13 + (COLS-16)/2));
        refresh();
        }
    *result_x = curr_x;
    *result_y = curr_y;
    return (command=='.');
    }

void describe_space(struct small_creature *c, long x, long y)
    {
    long foo, i;
    struct space_struct curr_space;
    struct clist *traverse;
    long printed;
    char message[256];
    if (can_see(c, x, y))
        {
        clear();
        game_variables->fs_messages_printed = 0;
        curr_space = current_level->space[x][y];
        find_string(curr_space.space_description, message);
        if ((message[0] >= 'a') && (message[0] <= 'z'))
            message[0] += 'A' - 'a';
        fs_pmsg(message);
        fs_pmsg("");
        printed = 0;
        if (curr_space.au > 0)
            {
#ifdef HAS_SNPRINTF
            snprintf(message, 255, "%ld gold piece", curr_space.au);
#else
            sprintf(message, "%ld gold piece", curr_space.au);
#endif
            message[255] = '\0';
            if (curr_space.au > 1)
                strncat(message, "s", max(0, 255 - strlen(message)));
            message[255] = '\0';
            strncat(message, ".", max(0, 255 - strlen(message)));
            message[255] = '\0';
            fs_pmsg(message);
            printed = 1;
            }
        if (curr_space.el > 0)
            {
#ifdef HAS_SNPRINTF
            snprintf(message, 255, "%ld electrum piece", curr_space.el);
#else
            sprintf(message, "%ld electrum piece", curr_space.el);
#endif
            message[255] = '\0';
            if (curr_space.el > 1)
                strncat(message, "s", max(0, 255 - strlen(message)));
            message[255] = '\0';
            strncat(message, ".", max(0, 255 - strlen(message)));
            message[255] = '\0';
            fs_pmsg(message);
            printed = 1;
            }
        if (curr_space.si > 0)
            {
#ifdef HAS_SNPRINTF
            snprintf(message, 255, "%ld silver piece", curr_space.si);
#else
            sprintf(message, "%ld silver piece", curr_space.si);
#endif
            message[255] = '\0';
            if (curr_space.si > 1)
                strncat(message, "s", max(0, 255 - strlen(message)));
            message[255] = '\0';
            strncat(message, ".", max(0, 255 - strlen(message)));
            message[255] = '\0';
            fs_pmsg(message);
            printed = 1;
            }
        if (curr_space.cu > 0)
            {
#ifdef HAS_SNPRINTF
            snprintf(message, 255, "%ld copper piece", curr_space.cu);
#else
            sprintf(message, "%ld copper piece", curr_space.cu);
#endif
            message[255] = '\0';
                if (curr_space.cu > 1)
            strncat(message, "s", max(0, 255 - strlen(message)));
            message[255] = '\0';
            strncat(message, ".", max(0, 255 - strlen(message)));
            message[255] = '\0';
            fs_pmsg(message);
            printed = 1;
            }
        if (printed)
            {
            fs_pmsg("");
            printed = 0;
            }
        if (!(curr_space.items==NULL))
            {
            print_ilist_internal(curr_space.items, 0, &foo, FULLSCREEN);
            printed = 1;
            }
        if (printed)
            {
            fs_pmsg("");
            printed = 0;
            }
        if (!(curr_space.creatures == NULL))
            {
            for(traverse = curr_space.creatures; traverse != NULL; traverse =
              traverse->next)
                {
        if (traverse->current != NULL)
        {
            describe_creature(traverse->current);
            fs_pmsg("");
        }
                }
            }
        if (flag_check(curr_space.flags, S_CLOAKED_BY_MIST))
            {
            fs_pmsg("A thick cloak of mist.");
            fs_pmsg("");
            }
        if (current_level->space[x][y].sign != ST_EMPTY)
            {
            fs_pmsg("You see a sign:");
            if (c != NULL)
                if (c->extended != NULL)
                    {
                    if (skill_check(c,
                      c->extended->skills->av[READING_WRITING], MODERATE,
                      SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE) > 0)
                        {
                        if (language_check(c, current_level->race))
                            {
                            find_string(current_level->space[x][y].sign,
                              message);
                            fs_pmsg(message);
                if (flag_check(current_level->space[x][y].flags,
                  OTHER_IS_LOCATION_TYPE))
                    for(i = 0; i < NUM_LOCATION_TYPES; ++i)
                    if (current_level->location[i].identity ==
                      current_level->space[x][y].other)
                        if (!(flag_check(current_level->location
                          [i].flags, LOCATION_KNOWN)))
                            {
                            flag_set(&current_level->location[i].flags,
                              LOCATION_KNOWN);
                            current_level->location[i].x = x;
                            current_level->location[i].y = y;
                            }
                            }
                        else
                            read_random_excerpt(FILE_TRANSLATION_FAILED);
                        }
                    else
                        fs_pmsg("(You can't make out what it says.)");
                    }
            fs_pmsg("");
            }
        fs_anykey();
        refresh_screen();
        }
    else if (can_see_badly(c, x, y))
        {
        curr_space = current_level->space[x][y];
        find_string(curr_space.space_description, message);
        if ((message[0] >= 'a') && (message[0] <= 'z'))
            message[0] += 'A' - 'a';
        pmsg(message);
        if ((!(curr_space.items==NULL)) || (!(curr_space.creatures==NULL)))
            pmsg("Something else; you can't tell what.");
        if (flag_check(curr_space.flags, S_CLOAKED_BY_MIST))
            {
            fs_pmsg("A thick cloak of mist.");
            fs_pmsg("");
            }
        }
    else
        pmsg("You can't see what's there.");
    }

void read_file(char *filename)
    {
    char real_filename[256], this_line[256], player_search_master[256],
      player_search_temp[256];
    struct read_file_text_struct *head, *tail, *current, *traverse;
    long my_continue, i;
    FILE *infile;
    head = tail = current = NULL;
    strncpy(real_filename, LIB_DIR, 255);
    real_filename[255] = '\0';
    strncat(real_filename, filename, max(0, 255 - strlen(real_filename)));
    player_search_master[0] = '\0';
    player_search_temp[0] = '\0';
    game_variables->fs_messages_printed=0;
    if (!(infile = fopen(real_filename, "r")))
        {
        perror("Attempt failed to open file:");
        perror(real_filename);
        }
    else
        {
        clear();
        my_continue = TRUE;
        while (my_continue)
            {
            my_continue = ((my_continue) && (fgets(this_line, 256, infile)));
            if (my_continue)
                {
                current = (struct read_file_text_struct *) calloc(1,
                  sizeof(struct read_file_text_struct));
                current->me = current;
                strncpy(current->text, this_line, 255);
                current->text[255] = '\0';
                current->next = NULL;
                current->prev = tail;
                if (tail != NULL)
                    tail->next = current;
                tail = current;
                if (head == NULL)
                    head = current;
                }
            }
        fclose(infile);
        current = head;
        my_continue = TRUE;
        if (head != NULL)
            while (my_continue)
                {
                if (current != NULL)
                    switch(rf_fs_pmsg(current->text))
                        {
                        case READ_FILE_CONTINUE: case READ_FILE_PAGE_DOWN:
                            current = current->next;
                        break;
                        case READ_FILE_HELP:
                            read_file(FILE_READ_FILE_HELP);
                            clear();
                            for(i = 0; i < LINES - 3; ++i)
                                if (current->prev != NULL)
                                    current = current->prev;
                        break;
                        case READ_FILE_PAGE_UP:
                            for(i = 0; i < 2 * (LINES - 2) - 1; ++i)
                                if (current->prev != NULL)
                                    current = current->prev;
                        break;
                        case READ_FILE_QUIT:
                            my_continue = FALSE;
                        break;
                        case READ_FILE_REFRESH:
                            for(i = 0; i < LINES - 3; ++i)
                                if (current->prev != NULL)
                                    current = current->prev;
                        break;
                        case READ_FILE_SEARCH_BACKWARDS:
                            clear();
                            rf_fs_pmsg("");
                            rf_fs_pmsg("Please enter the string you would like to search for.");
                            tgets(player_search_temp);
                            if (strlen(player_search_temp) == 0)
                                {
                                strncpy(player_search_temp,
                                  player_search_master, 255);
                                player_search_temp[255] = '\0';
                                }
                            for(i = 0; i < LINES - 3; ++i)
                                if (current->prev != NULL)
                                    current = current->prev;
                            for(traverse = current; (traverse != NULL) &&
                              (!(strcnt(traverse->text, player_search_temp)));
                                traverse = traverse->next);
                            if (traverse == NULL)
                                {
                                clear();
                                rf_fs_pmsg("Expression not found...");
                                fs_anykey();
                                }
                            else
                                current = traverse;
                        break;
                        case READ_FILE_SEARCH_FORWARDS:
                            clear();
                            rf_fs_pmsg("");
                            rf_fs_pmsg("Please enter the string you would like to search for.");
                            tgets(player_search_temp);
                            if (strlen(player_search_temp) == 0)
                                {
                                strncpy(player_search_temp,
                                  player_search_master, 255);
                                player_search_temp[255] = '\0';
                                }
                            for(traverse = current; (traverse != NULL) &&
                              (!(strcnt(traverse->text, player_search_temp)));
                                traverse = traverse->next);
                            if (traverse == NULL)
                                {
                                clear();
                                rf_fs_pmsg("Expression not found...");
                                fs_anykey();
                                for(i = 0; i < LINES - 3; ++i)
                                    if (current->prev != NULL)
                                        current = current->prev;
                                }
                            else
                                current = traverse;
                        break;
                        }
                else
                    my_continue = FALSE;
                }
        fs_anykey();
        clear();
        free_read_file_text_struct(head);
        if (!(flag_check(game_variables->flags, DURING_INIT)))
            refresh_screen();
        }
    }

/* To be called when creature takes an action which, on average, takes */
/* 'seconds' seconds to complete.  Accounts for creature's speed and haste */
/* value.  */

void take_time(struct small_creature *c, long time)
    {
    struct clist *traverse;
    long hostility;
    if (time < DAY)
        if ((c == player) && (time > 15L * SECOND))
            {
            hostility = FALSE;
            for(traverse = current_level->creatures; traverse != NULL;
              traverse = traverse->next)
            if (traverse->current != NULL)
                hostility |= is_hostile(traverse->current, c);
            if (hostility)
                c->nextaction += attribute_adjust(time, 0 - c->sp);
            else
                pass_time(attribute_adjust(time, 0 - c->sp));
            }
        else
            c->nextaction += attribute_adjust(time, 0 - c->sp);
    else perror("Error; take_time invoked for too great a time.");
    }

long upgrade_check(long power_level)
    {
    long i, failure;
    failure = 1;
    for(i=0;(i<power_level && failure);++i)
        failure = (failure && (trandom() % CHANCE_NEXT_LEVEL > 0));
    return (!(failure));
    }

long creature_epvalue(struct master_creature *c)
    {
    return c->po +
        c->wa +
        max(1, c->perm_attributes[AG]) +
        max(1, c->perm_attributes[AL]) +
        max(1, c->perm_attributes[CO]) +
        max(1, c->perm_attributes[DE]) +
        max(1, c->perm_attributes[SPEED]) +
        max(1, c->perm_attributes[ST]);
    }

void gain_experience(struct small_creature *earner, long units)
    {
    long units_gained, skill_chosen, garbage;
    struct small_creature *c;
    if (player != NULL && earner != player && is_following(earner, player))
        {
        c = player;
        }
    else
        {
        c = earner;
        }
    if (!(c->extended == NULL))
        {
        c->extended->partial_ep += units;
        if (c->extended->partial_ep >= UNITS_PER_EP)
            {
            units_gained = c->extended->partial_ep/UNITS_PER_EP;
            c->extended->partial_ep -= units_gained*UNITS_PER_EP;
            c->extended->unspent_ep += units_gained;
            c->extended->study_points += units_gained;
            if (c == player)
                pmsg("You gain experience...  (Hit 'A' to allocate your new experience.)");
            }
        if (c != player)
            {
            switch(trandom() % 3)
                {
                case 0: case 1:
                    switch(c->extended->profession)
                        {
                        case ACROBAT:
                            random_select(acrobat, &skill_chosen, &garbage);
                        break;
                        case ANIMAL_HANDLER:
                            random_select(animal_handler, &skill_chosen,
                              &garbage);
                        break;
                        case ARCHER:
                            random_select(archer, &skill_chosen, &garbage);
                        break;
                        case BARD:
                            random_select(bard, &skill_chosen, &garbage);
                        break;
                        case HERBALIST:
                            random_select(herbalist, &skill_chosen, &garbage);
                        break;
                        case HUNTER:
                            random_select(hunter, &skill_chosen, &garbage);
                        break;
                        case INTERPRETER:
                            random_select(interpreter, &skill_chosen,
                              &garbage);
                        break;
                        case JACK_OF_ALL_TRADES:
                            random_select(jack_of_all_trades, &skill_chosen,
                              &garbage);
                        break;
                        case MAN_AT_ARMS:
                            random_select(man_at_arms, &skill_chosen,
                              &garbage);
                        break;
                        case MEDIC:
                            random_select(medic, &skill_chosen, &garbage);
                        break;
                        case SCHOLAR:
                            random_select(scholar, &skill_chosen, &garbage);
                        break;
                        case SCOUT:
                            random_select(scout, &skill_chosen, &garbage);
                        break;
                        case WOODSMAN:
                            random_select(woodsman, &skill_chosen, &garbage);
                        break;
                        default:
                            random_select(skills_dist, &skill_chosen,
                              &garbage);
                        break;
                        }
                    switch(skill_chosen)
                        {
                        case ATTACK_HTH:
                            random_select(attack_hth_dist, &skill_chosen,
                              &garbage);
                        break;
                        case ATTACK_MSL:
                            random_select(attack_msl_dist, &skill_chosen,
                              &garbage);
                        break;
                        case FENCING:
                            random_select(fencing_dist, &skill_chosen,
                              &garbage);
                        break;
                        case MUSIC:
                            random_select(music_dist, &skill_chosen, &garbage);
                        break;
                        }
                default:
                    random_select(skills_dist, &skill_chosen, &garbage);
                break;
                }
            allocate_experience_internal(c, skill_chosen,
              c->extended->unspent_ep);
            }
        }
    }

void gain_ep(struct small_creature *c, long ep_gain)
    {
    gain_experience(c, ep_gain * UNITS_PER_EP);
    }

long ltrandom()
    {
    return trandom()*32768L + trandom();
    }

/* This returns TRUE if the damage kills the creature, FALSE otherwise.  We do
not invoke c_death() from within this function, in order to allow the calling
function to do other necessary operations (like award experience).  The flags
are creature flags.  */

long creature_damage(struct small_creature *c, long damage, long flags, long
  success)
    {
    char message[256], title[256];
    long result, exp_old_hv, exp_new_hv, exp_dam, local_damage;
    flag_reset(&c->flags, CONTROL_MOVE);
    flag_reset(&c->flags, SHIFT_MOVE);
    flag_reset(&c->flags, INACTIVE);
    if (flag_check(flags, CREATURE_DAMAGE_WAKE_UP))
        {
        if (flag_check(c->flags, SLEEPING))
            wake_up(c);
        c->nextaction = min(c->nextaction, game_variables->elapsed_units +
          SECOND);
        }
    result = FALSE;
    /* We exponentiate the damage the attacker does and the remaining */
    /* health of the creature, and subtract, then take the log.  We add 50 */
    /* to the values before exponentiation, so that the average case is */
    /* not all or nothing.  */
    if (status_check(c, STATUS_MIRACLE_AEGIS_OF_GOD) <= 0)
        {
        exp_old_hv = my_exp(50 + c->hv);
        exp_dam = trandom() % max(1, my_exp(50 + damage));
        /* We account for armor damage absorption */
        if (flag_check(flags, CREATURE_DAMAGE_ARMOR_ABSORBS))
            {
            if (flag_check(flags, CREATURE_DAMAGE_BLUNT_DAMAGE))
                exp_dam = (exp_dam * (100 - c->blunt))/100;
            else
                exp_dam = (exp_dam * (100 - c->sharp))/100;
            }
        else
                exp_dam = (exp_dam * 100)/100;
        exp_new_hv = exp_old_hv - exp_dam;
        if (exp_new_hv <= 0)
            result = TRUE;
        else
            {
        /* Decreased by one so that a minor wound does not automatically do one
         * damage. (Or tried that; the result seemed to neutralize far too many
         * wounds, so I changed back.)
            local_damage = 50 + c->hv - my_log(exp_new_hv);
         * As of 30 Mar 03, critical success does at minimum as many damage as
         * the levels of critical success, but otherwise this was as the
         * experimental version.
         */
            local_damage = max(success - 1, 49 + c->hv - my_log(exp_new_hv));
            if (c->extended == NULL)
                {
                c->att -= 3 * local_damage;
                c->def -= 2 * local_damage;
                c->dmg -= local_damage;
                c->hv -= local_damage;
                c->sp -= local_damage;
                c->wounded += local_damage;
                }
            else
                {
                c->extended->stati->time[STATUS_WOUNDED] += local_damage;

/* For the sake of simplicity, we assume that poison always causes half
   the damage of a venomous attack.  All damage caused by poison is
   healable by an antidote if treated quickly enough; half of the damage
   caused physically is quickly healable by the medical skill (the rest
   taking time and care to heal).  */

                if (flag_check(flags, HAS_VENOM))
                    {
                    c->extended->stati->time[STATUS_HEALABLE_ANTIDOTE] +=
                      (adjustment_info[STATUS_HEALABLE_ANTIDOTE].time_increment
                      * local_damage)/2;
                    c->extended->stati->time[STATUS_HEALABLE] +=
                      (adjustment_info[STATUS_HEALABLE].time_increment *
                      local_damage)/4;
                    }
                else
                    c->extended->stati->time[STATUS_HEALABLE] +=
                      (adjustment_info[STATUS_HEALABLE].time_increment *
                      damage)/2;
                calc_ready(c);
                }
            }
        return result;
        }
    else
        {
        if (c == player)
            pmsg("You are protected by the aegis of God!");
        else
            {
            find_title(c, title);
#ifdef HAS_SNPRINTF
            sprintf(message, "%s is protected by the aegis of God!", title);
#else
#endif
            message[255] = '\0';
            pmsg(message);
            }
        return 0;
        }
    }
